/*
    Copyright (c) 2016 digitalSTROM AG, Zurich, Switzerland

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/
#pragma once

#include <boost/asio/io_service.hpp>
#include <boost/asio/ip/tcp.hpp>
#include <boost/asio/steady_timer.hpp>
#include <boost/asio/streambuf.hpp>
#include <boost/signals2.hpp>
#include <boost/chrono.hpp>

namespace ds {
namespace vdsm {

/**
 * Reads and writes vdc frames from / to socket.
 *
 * Class provides asynchronous write api byt buffering output in write buffer.
 * If other side of socket does not read written data fast enough
 * and output buffer overflows, connections is dropped to recover.
 *
 * Class runs internal read cycle, and calls readFrame callback
 * for each parsed frame. There is no way to throttle read from socket.
 *
 * Api provided by this class is more push based compared to PULL asio api.
 * This class is not trying to support zero copy, etc.
 * It is rather trying to implement the business logic for frame handling
 * in VDC server / client with convenient API.
 *
 * It is safe to delete the object at any time.
 * Handlers queued in io_service that captured this may be invoked
 * after delete but they do nothing on error operation_aborted.
 * It is incorrect to delete socket passed to \ref start before
 * deleting this object. Call \ref close before deleting the socket.
 *
 * This class is usable for client as well as for server.
 */
class VdcFramer
{
    static constexpr std::size_t MaxMessageSize = 16384;
    static constexpr std::size_t MaxWriteBufferSize = 2 * MaxMessageSize;

public:
    VdcFramer(boost::asio::io_service& ioService);

    /**
     * Start new session. Signal closed is called once to notify end of session.
     *
     * Socket object must be valid till the end of session.
     * Note that closing socket does not close the session.
     */
    void start(boost::asio::ip::tcp::socket& socket);

    /** Close session. Does nothing if session is inactive. */
    void close();

    /** Called once after start() when session is closed. */
    boost::signals2::signal<void ()>& closed() { return m_closed; }

    void setReadTimeout(boost::chrono::milliseconds x) { m_readTimeout = x; }
    boost::chrono::milliseconds readTimeout() const { return m_readTimeout; }

    /**
     * Function falled when frame was read.
     * Passed data are valid only during the function call.
     *
     * Class internally runs the assio::read loop
     * and invokes callback for every parsed frame.
     */
    void setReadFrameFunction(std::function<void (const void *, size_t)>&& x)
    {
        m_readFrameFunction = std::move(x);
    }

    /** Writes massage as new frame (header + data). Reconnects if write buffer is exceeeded. */
    void asyncWriteFrame(const void *data, size_t size);

private:
    boost::asio::io_service& m_ioService;
    boost::asio::basic_waitable_timer<boost::chrono::steady_clock> m_readTimer;
    boost::asio::ip::tcp::socket* m_socket;
    boost::signals2::signal<void ()> m_closed;

    std::function<void (const void *, size_t)> m_readFrameFunction;

    //read state
    typedef uint16_t FrameHeaderType;
    FrameHeaderType m_readHeaderData;
    std::vector<char> m_readVector;

    //write state
    boost::asio::streambuf m_writeStreambuf;
    std::ostream m_writeStream;
    bool m_writeActive;
    boost::chrono::milliseconds m_readTimeout;

    void closeImpl();
    void asynWaitReadTimer();
    void asyncReadFrameHeader();
    void asyncReadFrameBody(FrameHeaderType header);
    void asyncWrite();
};

}} //namespace ds::vdsm
